from yankee.util import camelize


def do_nothing(obj):
    return obj


class Deserializer(object):
    data_key = None
    many = False

    class Meta:
        pass

    def __init__(self, data_key=None, required=False, output_style=None):
        if self.data_key is None:
            self.data_key = data_key
        self.required = required
        self.output_style = output_style
        self.bind()

    def bind(self, name=None, parent=None):
        self.name = name
        self.parent = parent
        if self.parent is not None:
            self.Meta = getattr(parent, "Meta", dict())
        self.output_name = camelize(self.name) if self.name else None
        if self.data_key == False:
            self.accessor = do_nothing
        else:
            self.accessor = self.make_accessor()

    def make_accessor(self):
        return do_nothing

    def post_load(self, output_data, **kwargs):
        return output_data

    def deserialize(self, obj):
        if self.many:
            if obj is None:
                return list()
            return self.accessor(obj)
        else:
            if obj is None:
                return None
            plucked_obj = self.accessor(obj)
            if isinstance(plucked_obj, list):
                if len(plucked_obj) > 1:
                    raise ValueError("Key Function returned multiple results!")
                elif not plucked_obj:
                    return self.post_load(None)
                else:
                    return self.post_load(plucked_obj[0])
            return self.post_load(plucked_obj)



from __future__ import annotations

import datetime

from dateutil.parser import parse as parse_dt, isoparse

from yankee.util import clean_whitespace, is_valid

from .deserializer import Deserializer
from .schema import Schema


class Field(Deserializer):
    def __init__(self, data_key=None, required=False, many=False):
        super().__init__(data_key=data_key, required=required)
        self.required = required

    def deserialize(self, obj):
        result = super().deserialize(obj)
        if result is None and self.required:
            raise ValueError(
                f"Field {self.name} is required! Key {self.key} not found in {obj}"
            )
        return result


class String(Field):
    def __init__(self, data_key=None, required=False, attr=None, formatter=None):
        super().__init__(data_key, required)
        self.formatter = formatter or clean_whitespace

    def deserialize(self, elem) -> "Optional[str]":
        elem = super().deserialize(elem)
        if elem is None:
            return None
        else:
            return self.formatter(self.to_string(elem))

    def to_string(self, elem):
        return str(elem)


class DateTime(String):
    def __init__(
        self, data_key=None, required=False, attr=None, formatter=None, dt_format=None
    ):
        super().__init__(data_key, required, attr, formatter)
        if dt_format:
            self.parse_date = lambda s: datetime.datetime.strptime(s, dt_format)

    def parse_date(self, text:str):
        try:
            return isoparse(text)
        except ValueError:
            return parse_dt(text)

    def deserialize(self, elem) -> "Optional[datetime.datetime]":
        string = super(DateTime, self).deserialize(elem)
        if not string:
            return None
        return self.parse_date(string)      



class Date(DateTime):
    def deserialize(self, elem) -> "Optional[datetime.date]":
        date_time = super().deserialize(elem)
        return date_time.date() if date_time else None


class Boolean(String):
    def __init__(
        self, *args, true_value="true", case_sensitive=False, allow_none=True, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.true_value = true_value
        self.case_sensitive = case_sensitive
        self.allow_none = allow_none
        if not self.case_sensitive:
            self.true_value = self.true_value.lower()

    def deserialize(self, elem) -> "Optional[bool]":
        string = super(Boolean, self).deserialize(elem)
        if string is None:
            return None if self.allow_none else False
        if not self.case_sensitive:
            string = string.lower()
        return string == self.true_value


class Float(String):
    def deserialize(self, elem) -> "Optional[float]":
        string = super(Float, self).deserialize(elem)
        return float(string) if string is not None else None


class Integer(String):
    def deserialize(self, elem) -> "Optional[int]":
        string = super(Integer, self).deserialize(elem)
        return int(string) if string is not None else None


class Exists(Field):
    def deserialize(self, elem) -> bool:
        obj = super(Exists, self).deserialize(elem)
        return obj is not None


class Const(Field):
    def __init__(self, const, *args, **kwargs):
        self.const = const

    def deserialize(self, elem) -> "Any":
        return self.const


# Multiple Value Field


class List(Field):
    many = True

    def __init__(self, item_schema, data_key=None, formatter=lambda l: l, **kwargs):
        self.formatter = formatter
        self.item_schema = item_schema
        if callable(self.item_schema):
            self.item_schema = item_schema()
        super().__init__(data_key=data_key, **kwargs)

    def bind(self, name=None, schema=None):
        super().bind(name, schema)
        self.item_schema.bind(None, schema)

    def deserialize(self, elem) -> "List":
        elements = super().deserialize(elem)
        if elements is None:
            return list()
        output = [self.item_schema.deserialize(e) for e in elements]
        return self.formatter([r for r in output if is_valid(r)])


# Schema-Like Fields


class Combine(Schema):
    """Can have fields like a schema that are then
    passed as an object to a combine function that
    transforms it to a single string value"""

    def bind(self, name=None, parent=None):
        super().bind(name, parent)
        for name, field in self.fields.items():
            field.bind(name, self)

    def combine_func(self, obj):
        raise NotImplementedError("Must be implemented in subclass")

    def deserialize(self, et_elem) -> "Optional[str]":
        obj = super().deserialize(et_elem)
        return self.combine_func(obj)


class Alternative(Schema):
    """There may be a piece of data that has different names
    in different contexts. This has fields like a schema, then
    passes as a value the first non-empty or non-null result"""

    def deserialize(self, et_elem):
        obj = super().deserialize(et_elem)
        return next((v for v in obj.values() if is_valid(v)), None)


class Zip(Schema):
    _list_field = List
    """Sometimes data is provided as a bunch of arrays, like:
    {
        "name": ["Peter", "Parker"],
        "age": [15, 25],
    }
    and we want to build out complete records from this data.
    This field performs that step:
    """

    def bind(self, name=None, parent=None):
        super().bind(name=name, parent=parent)
        list_fields = dict()
        if not hasattr(self, "_keys"):
            self._keys = {k: v.data_key for k, v in self.fields.items()}

        for k, v in self.fields.items():
            v.data_key = None
            list_field = self._list_field(v, data_key=self._keys[k])
            list_field.bind(k, self)
            list_fields[k] = list_field
        self.fields = list_fields

    def lists_to_records(self, obj):
        keys = tuple(obj.keys())
        values = tuple(obj.values())
        return [dict(zip(keys, v)) for v in zip(*values)]

    def deserialize(self, raw_obj) -> "Dict":
        obj = super().deserialize(raw_obj)
        return self.lists_to_records(obj)



import datetime

import lxml.etree as ET
import pytest

from yankee.xml import Schema, ZipSchema
from yankee.xml import fields as f

from .fields import *

test_doc = """
<testDoc>
    <string>Some String Data</string>
    <date_time>2021-05-04T12:05</date_time>
    <date>2021-05-04</date>
    <booleans>
        <bool>True</bool>
        <bool>true</bool>
        <bool>False</bool>
        <bool>false</bool>
    </booleans>
    <float>1.234</float>
    <int>23</int>
    <exists>Something</exists>
    <part1>George</part1>
    <part2>Burdell</part2>
    <random>Some data</random>
</testDoc>
""".strip()


tree = ET.fromstring(test_doc.encode())
class NameSchema(Combine):
    part1 = f.Str("./part1")
    part2 = f.Str("./part2")

    def combine_func(self, obj):
        return f"{obj['part1']} {obj['part2']}"
class ExampleSchema(Schema):
    string = f.Str("./string")
    date_time = f.DT("./date_time")
    date = f.Date("./date")
    booleans = f.List(Bool, "./booleans/bool")
    float = f.Float("./float")
    int = f.Int("./int")
    exists = f.Exists("./exists")
    does_not_exist = f.Exists("./does_not_exist")
    name = NameSchema()


def test_fields():
    d = ExampleSchema()
    data = d.deserialize(tree)
    assert data["string"] == "Some String Data"
    assert data["dateTime"] == datetime.datetime(2021, 5, 4, 12, 5)
    assert data["date"] == datetime.date(2021, 5, 4)
    assert data["booleans"] == [True, True, False, False]
    assert data["float"] - 1.234 < 0.001
    assert data["int"] == 23
    assert data["exists"] == True
    assert data["doesNotExist"] == False
    assert data["name"] == "George Burdell"

ns_test_doc = """
<?xml version='1.0' encoding='UTF-8'?>
<h:testDoc xmlns:h="http:www.w3.org/TR/html4/">
    <h:string>Some String Data</h:string>
    <h:date_time>2021-05-04T12:05</h:date_time>
    <h:date>2021-05-04</h:date>
    <h:booleans>
        <bool>True</bool>
        <bool>true</bool>
        <bool>False</bool>
        <bool>false</bool>
    </h:booleans>
</h:testDoc>
""".strip()

ns_tree = ET.fromstring(ns_test_doc.encode())

class NsSchema(Schema):
    class Meta():
        namespaces = {
            "h": "http:www.w3.org/TR/html4/"
        }
    string = f.Str("./h:string")
    date_time = f.DT("./h:date_time")
    date = f.Date("./h:date")
    booleans = f.List(Bool, "./h:booleans/bool")
   
def test_ns_fields():
    d = NsSchema()
    data = d.deserialize(ns_tree)
    assert data["string"] == "Some String Data"
    assert data["dateTime"] == datetime.datetime(2021, 5, 4, 12, 5)
    assert data["date"] == datetime.date(2021, 5, 4)
    assert data["booleans"] == [True, True, False, False]



from yankee.base import schema

from .fields import List
from .key import JsonMixin


class Schema(JsonMixin, schema.Schema):
    pass


class PolymorphicSchema(JsonMixin, schema.PolymorphicSchema):
    pass


class ZipSchema(JsonMixin, schema.ZipSchema):
    _list_field = List



from copy import deepcopy
from yankee.util import is_valid

from .deserializer import Deserializer


class Schema(Deserializer):
    class Meta:
        output_style = None

    def __init__(
        self,
        data_key=None,
        flatten=False,
        prefix=False,
        required=False,
        output_style=None,
    ):
        self.Meta.output_style = output_style
        self.flatten = flatten
        self.prefix = prefix
        super().__init__(data_key, required, output_style)

    def bind(self, name=None, parent=None):
        super().bind(name, parent)
        # Make sure that fields are grabbed from superclasses as well
        class_fields = list()
        for c in reversed(self.__class__.mro()):
            class_fields += [
                (k, v) for k, v in c.__dict__.items() if isinstance(v, Deserializer)
            ]
        self.fields = dict(class_fields)
        for name, field in self.fields.items():
            if self.prefix:
                field.bind(f"{self.name}_{name}", self)
            else:
                field.bind(name, self)

    def deserialize(self, raw_obj) -> "Dict":
        obj = super().deserialize(raw_obj)
        output = dict()
        for field in self.fields.values():
            value = field.deserialize(obj)
            # If there is no value, don't include anything in the output dictionary
            if not is_valid(value):
                if field.required == True:
                    return dict()
                continue
            # If the value isn't a dict, or there's not flatten directive, add and continue
            if not isinstance(value, dict) or not field.flatten:
                output[field.output_name] = value
                continue
            # Merge in flattened fields
            output.update(value)
        return output


class PolymorphicSchema(Schema):
    def bind(self, name=None):
        super().bind(self, name)
        for schema in self.schemas:
            schema.bind(name)

    def choose_schema(self, obj):
        raise NotImplementedError("Must be implemented in subclass!")

    def deserialize(self, raw_obj) -> "Dict":
        # Get the key one time only, rather than
        # on both deserializing the selector obj
        # and the final output
        obj = super().deserialize(obj)
        schema = self.choose_schema(obj)
        return schema.deserialize(raw_obj)


class ZipSchema(Schema):
    _list_field = None
    """Sometimes data is provided as a bunch of arrays, like:
    {
        "name": ["Peter", "Parker"],
        "age": [15, 25],
    }
    and we want to build out complete records from this data.
    This field performs that step:
    """

    def bind(self, name=None, parent=None):
        super().bind(name=name, parent=parent)
        from yankee.xml import Deserializer as XmlDeserializer
        if not hasattr(self, "_keys"):
            self._keys = list()
            for f in self.fields.values():
                generic_f = XmlDeserializer(f.data_key, f.required)
                generic_f.bind(f.name, parent)
                self._keys.append(generic_f)

    def lists_to_records(self, obj):
        keys = tuple(obj.keys())
        values = tuple(obj.values())
        return [dict(zip(keys, v)) for v in zip(*values)]

    def deserialize(self, raw_obj) -> "Dict":
        elements = [f.deserialize(raw_obj) for f in self._keys]
        import lxml.etree as ET
        docs = list(map(list, zip(*elements)))

        obj = super().deserialize(raw_obj)
        return self.lists_to_records(obj)



from lxml.etree import XPath


def do_nothing(obj):
    return obj


class XmlMixin(object):
    def make_accessor(self):
        namespaces = getattr(self.Meta, "namespaces", dict())
        if self.data_key is None:
            return do_nothing
        if self.data_key and self.many == False:
            return XPath(f"({self.data_key})[1]", namespaces=namespaces)
        else:
            return XPath(self.data_key, namespaces=namespaces)

    def to_string(self, elem):
        return "".join(elem.itertext())
